from supervisely.nn.inference.inference import Inference
from supervisely.nn.inference.gui.gui import InferenceGUI, BaseInferenceGUI
from supervisely.nn.inference.instance_segmentation.instance_segmentation import (
    InstanceSegmentation,
)
from supervisely.nn.inference.object_detection.object_detection import ObjectDetection
from supervisely.nn.inference.semantic_segmentation.semantic_segmentation import (
    SemanticSegmentation,
)
from supervisely.nn.inference.pose_estimation.pose_estimation import PoseEstimation
from supervisely.nn.inference.session import Session, SessionJSON



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
import supervisely as sly
from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import Widget
from supervisely.api.api import Api


class ModelInfo(Widget):
    def __init__(
        self,
        session_id: int = None,
        team_id: int = None,
        widget_id: str = None,
    ):
        self._api = Api()
        self._session_id = session_id
        self._team_id = team_id

        if self._team_id is None:
            self._team_id = sly.env.team_id()

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        data = {}
        data["teamId"] = self._team_id
        if self._session_id is not None:
            data["model_info"] = self._api.task.send_request(
                self._session_id, "get_session_info", data={}
            )
        else:
            data["model_info"] = None
        if self._session_id is not None:
            data["model_connected"] = True
        else:
            data["model_connected"] = False
        return data

    def get_json_state(self):
        state = {}
        state["sessionId"] = self._session_id
        return state

    def set_session_id(self, session_id):
        self._session_id = session_id
        self.update_data()
        self.update_state()
        DataJson().send_changes()
        StateJson().send_changes()



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from typing import Union, List


class Slider(Widget):
    def __init__(
            self,
            value: Union[int, List[int]] = 0,
            min: int = 0,
            max: int = 100,
            step: int = 1,
            show_input: bool = False,
            show_input_controls: bool = False,
            show_stops: bool = False,
            show_tooltip: bool = True,
            range: bool = False,  # requires value to be List[int, int]
            vertical: bool = False,
            height: int = None,
            widget_id: str = None
    ):
        self._value = value
        self._min = min
        self._max = max
        self._step = step
        self._show_input = show_input
        self._show_input_controls = show_input_controls if show_input else False
        self._show_stops = show_stops
        self._show_tooltip = show_tooltip
        self._range = False if show_input else range
        self._vertical = vertical
        self._height = f"{height}px" if vertical else None

        self._validate_min_max(min, max)
        self._validate_default_value(value)

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _validate_min_max(self, min_val: int, max_val: int):
        if min_val > max_val:
            raise ValueError(f"Minimum value: '{min_val}' can't be bigger than maximum value: '{max_val}'")

    def _validate_default_value(self, value: Union[int, List[int]]):
        if isinstance(value, int):
            if self._range:
                raise ValueError(f"value = '{value}', should be 'list' if range is True")
            if value < self._min:
                self._value = self._min
            elif value > self._max:
                self._value = self._max
        elif isinstance(value, list):
            if not self._range:
                raise ValueError(f"value = '{value}', should be 'int' if range is False")
            if value[0] < self._min and value[1] > self._max:
                self._value = self._min
                self._value = self._max
            elif value[0] < self._min and value[1] < self._min:
                self._value[0] = self._min
                self._value[1] = self._min
            elif value[0] > self._max and value[1] > self._max:
                self._value[0] = self._max
                self._value[1] = self._max
        else:
            raise ValueError(f"value = '{value}', should be 'int' or 'list', not '{type(value)}'")

    def get_json_data(self):
        return {
            "min": self._min,
            "max": self._max,
            "step": self._step,
            "showInput": self._show_input,
            "showInputControls": self._show_input_controls,
            "showStops": self._show_stops,
            "showTooltip": self._show_tooltip,
            "range": self._range,
            "vertical": self._vertical,
            "height": self._height
        }

    def get_json_state(self):
        return {"value": self._value}

    def set_value(self, value: Union[int, List[int]]):
        self._validate_default_value(value)
        StateJson()[self.widget_id]["value"] = value
        StateJson().send_changes()

    def get_value(self):
        return StateJson()[self.widget_id]["value"]

    def set_min(self, value: int):
        self._validate_min_max(value, self.get_max())
        DataJson()[self.widget_id]["min"] = value
        DataJson().send_changes()

    def get_min(self):
        return DataJson()[self.widget_id]["min"]

    def set_max(self, value: int):
        self._validate_min_max(self.get_min(), value)
        DataJson()[self.widget_id]["max"] = value
        DataJson().send_changes()

    def get_max(self):
        return DataJson()[self.widget_id]["max"]

    def set_step(self, value: int):
        DataJson()[self.widget_id]["step"] = value
        DataJson().send_changes()

    def get_step(self):
        return DataJson()[self.widget_id]["step"]

    def is_input_enabled(self):
        return DataJson()[self.widget_id]["showInput"]

    def show_input(self):
        DataJson()[self.widget_id]["showInput"] = True
        DataJson().send_changes()

    def hide_input(self):
        DataJson()[self.widget_id]["showInput"] = False
        DataJson().send_changes()

    def is_input_controls_enabled(self):
        return DataJson()[self.widget_id]["showInputControls"]

    def show_input_controls(self):
        DataJson()[self.widget_id]["showInputControls"] = True
        DataJson().send_changes()

    def hide_input_controls(self):
        DataJson()[self.widget_id]["showInputControls"] = False
        DataJson().send_changes()

    def is_step_enabled(self):
        return DataJson()[self.widget_id]["showStops"]

    def show_steps(self):
        DataJson()[self.widget_id]["showStops"] = True
        DataJson().send_changes()

    def hide_steps(self):
        DataJson()[self.widget_id]["showStops"] = False
        DataJson().send_changes()

    def is_tooltip_enabled(self):
        return DataJson()[self.widget_id]["showTooltip"]

    def show_tooltip(self):
        DataJson()[self.widget_id]["showTooltip"] = True
        DataJson().send_changes()

    def hide_tooltip(self):
        DataJson()[self.widget_id]["showTooltip"] = False
        DataJson().send_changes()



import fastapi
import pandas as pd
from varname import varname
import numpy as np
import copy
from typing import Any
import traceback

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget

from supervisely.sly_logger import logger


class PackerUnpacker:
    SUPPORTED_TYPES = tuple([dict, pd.DataFrame])

    @staticmethod
    def validate_sizes(unpacked_data):

        if len(unpacked_data["data"]) != len(unpacked_data["classes"]):
            raise ValueError(
                "Sizes mismatch:\n"
                f'number of rows ({len(unpacked_data["data"])}) != number of columns ({len(unpacked_data["classes"])})'
            )

        for row in unpacked_data["data"]:
            if len(row) != len(unpacked_data["classes"]):
                raise ValueError(
                    "Sizes mismatch:\n"
                    f'{len(row)} != {len(unpacked_data["classes"])}\n'
                    f"{row}\n"
                    f'{unpacked_data["classes"]}'
                )

    @staticmethod
    def unpack_data(data, unpacker_cb, validate_sizes=True):
        unpacked_data = unpacker_cb(data)
        if validate_sizes:
            PackerUnpacker.validate_sizes(unpacked_data)
        return unpacked_data

    @staticmethod
    def pack_data(data, packer_cb):
        packed_data = packer_cb(data)
        return packed_data

    @staticmethod
    def dict_unpacker(data: dict):
        formatted_rows = []
        for origin_row in data["data"]:
            formatted_rows.append([{"value": element} for element in origin_row])

        unpacked_data = {"classes": data["columns"], "data": formatted_rows}

        return unpacked_data

    @staticmethod
    def pandas_unpacker(data: pd.DataFrame):
        data = data.where(pd.notnull(data), None)
        data = data.astype(object).replace(np.nan, "None")

        formatted_rows = []
        for origin_row in list(data.values):
            formatted_rows.append([{"value": element} for element in origin_row])

        unpacked_data = {"classes": data.columns.to_list(), "data": formatted_rows}
        return unpacked_data

    @staticmethod
    def dict_packer(data):
        unformatted_rows = []

        for origin_row in data["data"]:
            unformatted_rows.append([element["value"] for element in origin_row])

        packed_data = {
            "data": unformatted_rows,
            "columns": data["classes"],
        }
        return packed_data

    @staticmethod
    def pandas_packer(data):
        unformatted_rows = []

        for origin_row in data["data"]:
            unformatted_rows.append([element["value"] for element in origin_row])

        packed_data = pd.DataFrame(data=unformatted_rows, columns=data["classes"])
        return packed_data


DATATYPE_TO_PACKER = {pd.DataFrame: PackerUnpacker.pandas_packer, dict: PackerUnpacker.dict_packer}

DATATYPE_TO_UNPACKER = {
    pd.DataFrame: PackerUnpacker.pandas_unpacker,
    dict: PackerUnpacker.dict_unpacker,
}


class ConfusionMatrix(Widget):
    class Routes:
        CELL_CLICKED = "cell_clicked_cb"

    class ClickedDataPoint:
        def __init__(
            self,
            column_name: str,
            column_index: int,
            row_name: str,
            row_index: int,
            cell_value: Any,
        ):
            self.column_index = column_index
            self.column_name = column_name
            self.row_index = row_index
            self.row_name = row_name
            self.cell_value = cell_value

    def __init__(
        self,
        data: list = None,
        columns: list = None,
        x_label: str = "Predicted Values",
        y_label: str = "Actual Values",
        widget_id: str = None,
    ):
        """
        :param data: Data of table in different formats:
        1. Pandas Dataframe \n
                            2. Python dict with structure {
                                        'columns': ['col_name_1', 'col_name_2', ...],
                                        'data': [
                                                            ['row_1_column_1', 'row_1_column_2', ...],
                                                            ['row_2_column_1', 'row_2_column_2', ...],
                                                            ...
                                                          ]
                                      }
        """
        self._supported_types = PackerUnpacker.SUPPORTED_TYPES

        self._parsed_data = None
        self._parsed_data_with_totals = {}
        self._data_type = None

        self._update_matrix_data(input_data=pd.DataFrame(data=data, columns=columns))

        self.x_label = x_label
        self.y_label = y_label

        self._loading = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "matrix_data": self._parsed_data_with_totals,
            "matrix_options": {
                "selectable": True,
                "horizontalLabel": self.x_label,
                "verticalLabel": self.y_label,
            },
            "loading": self._loading,
        }

    def get_json_state(self):
        return {"selected_row": {}}

    def _update_matrix_data(self, input_data):
        if input_data is not None:
            self._parsed_data = copy.deepcopy(self.get_unpacked_data(input_data=input_data))
            self._data_type = type(input_data)
        else:
            self._parsed_data = {"classes": [], "data": []}
            self._data_type = dict
        self._calculate_totals()

    def _calculate_totals(self):
        matrix_data = []
        for origin_row in self._parsed_data["data"]:
            matrix_data.append([element["value"] for element in origin_row])

        totals_by_rows = np.asarray(["-" for _ in matrix_data]).reshape(-1, 1)
        totals_by_columns = np.asarray([["-" for _ in matrix_data]])

        try:
            matrix_data = np.matrix(matrix_data).astype(float)
            totals_by_rows = np.sum(matrix_data, axis=1).round(2)
            totals_by_columns = np.sum(matrix_data, axis=0).round(2)

            self._parsed_data_with_totals.update(self._calculate_max_values(matrix_data))

        except Exception as ex:
            logger.warning(f"Cannot calculate totals for matrix ({self.__class__.__name__}): {ex}")

        totals_by_columns = np.hstack([totals_by_columns, [[None]]])

        matrix_data = np.hstack([matrix_data, totals_by_rows])
        matrix_data = np.vstack([matrix_data, totals_by_columns])

        self._parsed_data_with_totals.update(
            self.get_unpacked_data(
                input_data={"columns": self._parsed_data["classes"], "data": matrix_data.tolist()},
                validate_sizes=False,
            )
        )

    def _get_packed_data(self, input_data, data_type):
        return PackerUnpacker.pack_data(data=input_data, packer_cb=DATATYPE_TO_PACKER[data_type])

    def get_unpacked_data(self, input_data, validate_sizes=True):
        input_data_type = type(input_data)

        if input_data_type not in self._supported_types:
            raise TypeError(
                f"Cannot parse input data, please use one of supported datatypes: {self._supported_types}\n"
                """
                            1. Pandas Dataframe \n
                            2. Python dict with structure {
                                        'columns': ['col_name_1', 'col_name_2', ...],
                                        'data': [
                                                            ['row_1_column_1', 'row_1_column_2', ...],
                                                            ['row_2_column_1', 'row_2_column_2', ...],
                                                            ...
                                                          ]
                                      }
                            """
            )

        return PackerUnpacker.unpack_data(
            data=input_data,
            unpacker_cb=DATATYPE_TO_UNPACKER[input_data_type],
            validate_sizes=validate_sizes,
        )

    def to_json(self) -> dict:
        return self._get_packed_data(self._parsed_data, dict)

    def to_pandas(self) -> pd.DataFrame:
        return self._get_packed_data(self._parsed_data, pd.DataFrame)

    def read_json(self, value: dict):
        self._update_matrix_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data

    def read_pandas(self, value: pd.DataFrame):
        self._update_matrix_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data

    def get_selected_cell(self, state):
        row_index = state[self.widget_id]["selected_row"].get("row")
        col_index = state[self.widget_id]["selected_row"].get("col")

        row_data = None
        cell_data = None

        if row_index is not None and col_index is not None:
            row_data = [element["value"] for element in self._parsed_data["data"][row_index]]
            cell_data = {
                "row_name": self._parsed_data["classes"][row_index],
                "col_name": self._parsed_data["classes"][col_index],
                "value": row_data[col_index],
            }

        return {
            "row_index": row_index,
            "col_index": col_index,
            "row_data": row_data,
            "cell_data": cell_data,
        }

    def _calculate_max_values(self, matrix_data):
        return {
            "diagonalMax": float(max(np.diagonal(matrix_data))),
            "maxValue": float(matrix_data.max()),
        }

    @property
    def loading(self):
        return self._loading

    @loading.setter
    def loading(self, value: bool):
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading

    def click(self, func):
        route_path = self.get_route_path(ConfusionMatrix.Routes.CELL_CLICKED)
        server = self._sly_app.get_server()

        @server.post(route_path)
        def _click():
            try:
                value_dict = self.get_selected_cell(StateJson())
                if value_dict is None:
                    return
                datapoint = ConfusionMatrix.ClickedDataPoint(
                    column_name=value_dict["cell_data"]["col_name"],
                    column_index=value_dict["col_index"],
                    row_name=value_dict["cell_data"]["row_name"],
                    row_index=value_dict["row_index"],
                    cell_value=value_dict["cell_data"]["value"],
                )
                func(datapoint)
            except Exception as e:
                logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
                raise e

        return _click



from supervisely.app.widgets import Widget
from supervisely.app.content import StateJson, DataJson


class LinePlot(Widget):
    def __init__(
        self,
        title: str,
        series: list = [],
        smoothing_weight: int = 0, 
        group_key: str = None, 
        show_legend: bool = True, 
        decimals_in_float: int = 2, 
        xaxis_decimals_in_float: int = None, 
        yaxis_interval: list = None, 
        widget_id = None,
        yaxis_autorescale: bool = True,  # issue in apex, need to refresh page
    ):
        self._title = title
        self._series = series
        self._smoothing_weight = smoothing_weight
        self._group_key = group_key
        self._show_legend = show_legend
        self._decimals_in_float = decimals_in_float
        self._xaxis_decimals_in_float = xaxis_decimals_in_float
        self._yaxis_interval = yaxis_interval
        self._options = {
            "title": self._title,
            "smoothingWeight": self._smoothing_weight,
            "groupKey": self._group_key,
            "showLegend": self._show_legend,
            "decimalsInFloat": self._decimals_in_float,
            "xaxisDecimalsInFloat": self._xaxis_decimals_in_float,
            "yaxisInterval": self._yaxis_interval
        }
        self._yaxis_autorescale = yaxis_autorescale
        self._ymin = 0
        self._ymax = 10
        super(LinePlot, self).__init__(widget_id=widget_id, file_path=__file__)
        self.update_y_range(self._ymin, self._ymax)

    def get_json_data(self):
        return {
            "title": self._title,
            "series": self._series,
            "options": self._options,
            "ymin": self._ymin,
            "ymax": self._ymax,
        }

    def get_json_state(self):
        return None
    
    def update_y_range(self, ymin: int, ymax: int, send_changes=True):
        self._ymin = min(self._ymin, ymin)
        self._ymax = max(self._ymax, ymax)
        if self._yaxis_autorescale is False:
            self._options["yaxis"][0]["min"] = self._ymin
            self._options["yaxis"][0]["max"] = self._ymax

    def add_series(self, name: str, x: list, y: list, send_changes: bool = True):
        assert len(x) == len(y), ValueError(f"Lists x and y have different lenght, {len(x)} != {len(y)}")

        data = [{"x": px, "y": py} for px, py in zip(x, y)]
        series = {"name": name, "data": data}
        self._series.append(series)
        
        if len(y) > 0:
            self.update_y_range(min(y), max(y))

        DataJson()[self.widget_id]['series'] = self._series
        if send_changes:
            DataJson().send_changes()

    def add_series_batch(self, series: list):
        for serie in series:
            name = serie["name"]
            x = serie["x"]
            y = serie["y"]
            self.add_series(name, x, y, send_changes=False)
        DataJson().send_changes()

    def add_to_series(self, name_or_id: str or int, data: dict):
        if isinstance(name_or_id, int):
            series_id = name_or_id
        else:
            series_id, _ = self.get_series_by_name(name_or_id)
        self._series[series_id]['data'] +=  data
        DataJson()[self.widget_id]['series'] = self._series
        DataJson().send_changes()

    def get_series_by_name(self, name):
        series_list = DataJson()[self.widget_id]['series']
        series_id, series_data = next(((i, series) for i, series in enumerate(series_list) if series['name'] == name), (None, None))
        # assert series_id is not None, KeyError("Series with name: {name} doesn't exists.")
        return series_id, series_data


# coding: utf-8
from __future__ import annotations
from typing import List, Dict, Optional, Iterator
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject
from supervisely.project.project_meta import ProjectMeta
from supervisely.video_annotation.key_id_map import KeyIdMap

class PointcloudObjectCollection(VideoObjectCollection):
    '''
    Collection that stores PointcloudObject instances.
    '''
    item_type = PointcloudObject

    def __iter__(self) -> Iterator[PointcloudObject]:
        return next(self)

    @classmethod
    def from_json(
        cls, 
        data: List[Dict], 
        project_meta: ProjectMeta, 
        key_id_map: Optional[KeyIdMap]=None
    ) -> PointcloudObjectCollection:
        return super().from_json(data, project_meta, key_id_map=key_id_map)